

import math

import numpy as np
import statsmodels.api as sm
import scipy.stats
import matplotlib.pyplot as plt

try:
	import piecewise_regression.r_squared_calc as r_squared_calc
	from piecewise_regression.data_validation import validate_positive_number, validate_boolean, validate_positive_integer, validate_list_of_numbers, validate_non_negative_integer
	from piecewise_regression.main import Fit
except:
	import r_squared_calc
	from data_validation import validate_positive_number, validate_boolean, validate_positive_integer, validate_list_of_numbers, validate_non_negative_integer
	from main import Fit

class ModelSelection:
	"""
	Experimental - uses simple BIC based on simple linear model.
	"""

	def __init__(self, xx, yy, max_breakpoints=10, n_boot=20, max_iterations=30, tolerance=10**-5,
		min_distance_between_breakpoints=0.01, min_distance_to_edge=0.02, verbose=True):


		self.models = []
		self.full_models_data = []
		
		self.stop = False

		if verbose==True:
			print("Running fit with n_breakpoint = 0 . . ")

		self.no_breakpoint_fit(xx, yy)

		for k in range(1, max_breakpoints+1):
			if verbose==True:
				print("Running fit with n_breakpoint = {} . . ".format(k))
			bootstrapped_fit = Fit(xx, yy, n_breakpoints=k, verbose=False, 
				n_boot=n_boot, max_iterations=max_iterations, tolerance=tolerance,
				min_distance_between_breakpoints=min_distance_between_breakpoints, min_distance_to_edge=min_distance_to_edge)
			fit_summary = bootstrapped_fit.get_results()
			fit_summary["n_breakpoints"] = k
			self.models.append(fit_summary)
			self.full_models_data.append(bootstrapped_fit)

		self.summary()


	def summary(self):
		
		header = "\n{:^70}\n".format("Breakpoint Model Comparision Results")

		line_length=100
		double_line = "=" * line_length + "\n"
		single_line = "-" * line_length + "\n"

		table_header_template = "{:<15} {:>12} {:>12} {:>12} \n"
		table_header = table_header_template.format("n_breakpoints", "BIC", "converged", "RSS")		
		table_row_template = "{:<15} {:>12.5} {:>12} {:>12.5} \n"

		table_contents = header
		table_contents += double_line

		table_contents += table_header
		table_contents += single_line

		for model in self.models:

			if model["converged"]:
				model_row = table_row_template.format(model["n_breakpoints"], model["bic"], str(model["converged"]), model["rss"])
			else:
				model_row = table_row_template.format(model["n_breakpoints"], "", str(model["converged"]), "")

			table_contents += model_row

		print(table_contents)

		print("Minimum BIC (Bayesian Information Criterion) suggests the best model")


	def no_breakpoint_fit(self, xx, yy):

		Z = np.array([xx])
		Z = Z.T
		Z = sm.add_constant(Z, has_constant='add')
		# Basic OLS fit
		results = sm.OLS(endog=np.array(yy), exog=Z).fit()

		# get the predicted values 
		ff = [(results.params[0] + results.params[1] * x) for x in xx]

		# Get Rss
		rss, tss, r_2, adjusted_r_2 = r_squared_calc.get_r_squared(yy, ff, n_params=2)

		# Calcualte BIC
		n = len(xx) # No. data points
		k =  2 # No. parameters
		bic = n * np.log(rss/n) + k * np.log(n)

		fit_data = {
			"bic":bic,
			"n_breakpoints":0,
			"estimates":{},
			"converged":True,
			"rss":rss
		}

		fit_data["estimates"]["const"] = results.params[0]
		fit_data["estimates"]["alpha1"] = results.params[1]

		self.models.append(fit_data)
		self.full_models_data.append(fit_data)


if __name__=="__main__":

	#np.random.seed(2)
	pass




import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
import math

import os, sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))

from piecewise_regression import Fit
from piecewise_regression import ModelSelection


def on_data_1():

	alpha = -4
	beta_1 = -2
	intercept = 100
	breakpoint_1 = 7

	n_points = 200

	xx = np.linspace(0, 20, n_points)
	yy = intercept + alpha*xx + beta_1 * np.maximum(xx - breakpoint_1, 0) + np.random.normal(size=n_points)

		

	pw_fit = Fit(xx, yy, start_values=[5])


	print("p-value is ", pw_fit.davies)

	pw_results = pw_fit.get_results()
	pw_estimates = pw_results["estimates"]
	print(pw_results)

	print(pw_estimates)

	pw_bootstrap_history = pw_fit.bootstrap_history
	print(pw_bootstrap_history)



	#print(bp_fit.breakpoint_history)

	#bp_fit.plot_data()
	#plt.show()


def on_data_1b():

	alpha = -4
	beta_1 = -4
	beta_2 = 4
	intercept = 100
	breakpoint_1 = 7
	breakpoint_2 = 12

	n_points = 200

	xx = np.linspace(0, 20, n_points)

	yy = intercept + alpha*xx + beta_1 * np.maximum(xx - breakpoint_1, 0) + beta_2 * np.maximum(xx-breakpoint_2, 0)  + np.random.normal(size=n_points)


	bp_fit = Fit(xx, yy, start_values=[5, 10])

	bp_fit.summary()

	bp_fit.plot_best_muggeo_breakpoint_history()
	plt.show()

	bp_fit.plot_data()
	bp_fit.plot_fit(color="red", linewidth=4)
	bp_fit.plot_breakpoints()
	bp_fit.plot_breakpoint_confidence_intervals()
	plt.show()



def on_data_1c():

	alpha = -4
	beta_1 = -2
	beta_2 = 4
	beta_3 = 1
	intercept = 100
	breakpoint_1 = 7
	breakpoint_2 = 13
	breakpoint_3 = 14

	n_points = 200

	xx = np.linspace(0, 20, n_points)

	yy = intercept + alpha*xx 
	yy += beta_1 * np.maximum(xx - breakpoint_1, 0) 
	yy += beta_2 * np.maximum(xx - breakpoint_2, 0)  
	yy += beta_3 * np.maximum(xx - breakpoint_3, 0)
	yy += np.random.normal(size=n_points)


	bp_fit = Fit(xx, yy, start_values=[5, 10, 16])

	bp_fit.summary()


	
	bp_fit.plot_data()
	bp_fit.plot_fit(color="red", linewidth=4)
	bp_fit.plot_breakpoints()
	bp_fit.plot_breakpoint_confidence_intervals()


	print("The fit data: ", bp_fit.__dict__)


	plt.show()
	plt.close()

	bp_fit.plot_best_muggeo_breakpoint_history()
	plt.legend()
	plt.show()
	plt.close()

	bp_fit.plot_bootstrap_restarting_history()
	plt.legend()
	plt.show()
	plt.close()



def model_selection_1():

	alpha = -4
	beta_1 = -2
	intercept = 100
	breakpoint_1 = 17

	n_points = 100

	xx = np.linspace(10, 30, n_points)
	yy = intercept + alpha*xx + beta_1 * np.maximum(xx - breakpoint_1, 0) + np.random.normal(size=n_points)
	#yy = intercept + alpha*xx + beta_1 + np.random.normal(size=n_points)
	

	xx_str = [str(x) for x in xx]
	yy_str = [str(y) for y in yy]	

	ms = ModelSelection(xx, yy, max_breakpoints=6)


def model_selection_2():


	alpha = -4
	beta_1 = -4
	beta_2 = 4
	intercept = 100
	breakpoint_1 = 7
	breakpoint_2 = 12

	n_points = 200

	xx = np.linspace(0, 20, n_points)

	yy = intercept + alpha*xx + beta_1 * np.maximum(xx - breakpoint_1, 0) + beta_2 * np.maximum(xx-breakpoint_2, 0)  + np.random.normal(size=n_points)


	ms = ModelSelection(xx, yy)


def fit_3_check_this_makes_sense():

	np.random.seed(0)

	alpha = 10
	beta_1 = -8
	beta_2 = -6
	beta_3 = 10
	intercept = 100
	breakpoint_1 = 7
	breakpoint_2 = 10
	breakpoint_3 = 14

	n_points = 200

	xx = np.linspace(0, 20, n_points)

	yy = intercept + alpha*xx 
	yy += beta_1 * np.maximum(xx - breakpoint_1, 0) 
	yy += beta_2 * np.maximum(xx - breakpoint_2, 0)  
	yy += beta_3 * np.maximum(xx - breakpoint_3, 0)
	yy += np.random.normal(size=n_points)


	pr = Fit(xx, yy, n_breakpoints=2)
	pr.plot()
	plt.show()

	pr3 = Fit(xx, yy, n_breakpoints=3)
	pr3.plot()
	plt.show()

	pr4 = Fit(xx, yy, n_breakpoints=4)
	pr4.plot()
	plt.show()

	ms = ModelSelection(xx, yy, max_breakpoints=6)


def fit_with_initally_diverging():
	np.random.seed(2)

	alpha = 10
	beta_1 = -8
	beta_2 = 3
	beta_3 = 10
	intercept = 100
	breakpoint_1 = 7
	breakpoint_2 = 10
	breakpoint_3 = 14

	n_points = 200

	xx = np.linspace(0, 20, n_points)

	yy = intercept + alpha*xx 
	yy += beta_1 * np.maximum(xx - breakpoint_1, 0) 
	yy += beta_2 * np.maximum(xx - breakpoint_2, 0)  
	yy += beta_3 * np.maximum(xx - breakpoint_3, 0)
	yy += np.random.normal(size=n_points)


	pr = Fit(xx, yy, n_breakpoints=2)
	print(pr.summary)

def fit_with_initially_diverging_start_values():

	np.random.seed(0)

	alpha = 10
	beta_1 = -8
	beta_2 = 3
	beta_3 = 10
	intercept = 100
	breakpoint_1 = 7
	breakpoint_2 = 10
	breakpoint_3 = 14

	n_points = 200

	xx = np.linspace(0, 20, n_points)

	yy = intercept + alpha*xx 
	yy += beta_1 * np.maximum(xx - breakpoint_1, 0) 
	yy += beta_2 * np.maximum(xx - breakpoint_2, 0)  
	yy += beta_3 * np.maximum(xx - breakpoint_3, 0)
	yy += np.random.normal(size=n_points)


	pr = Fit(xx, yy, start_values=[2.15646833, 0.98300926], n_boot=20)
	pr.summary()

def fit_with_initially_diverging_start_values_b():

	np.random.seed(0)

	alpha = 10
	beta_1 = -8
	beta_2 = 3
	beta_3 = 10
	intercept = 100
	breakpoint_1 = 7
	breakpoint_2 = 10
	breakpoint_3 = 14

	n_points = 200

	xx = np.linspace(0, 20, n_points)

	yy = intercept + alpha*xx 
	yy += beta_1 * np.maximum(xx - breakpoint_1, 0) 
	yy += beta_2 * np.maximum(xx - breakpoint_2, 0)  
	yy += beta_3 * np.maximum(xx - breakpoint_3, 0)
	yy += np.random.normal(size=n_points)


	pr = Fit(xx, yy, start_values=[1.2, 0.53], n_boot=25)
	pr.summary()

def fit_with_straight_line():

	np.random.seed(0)

	alpha = 10
	intercept = 100

	n_points = 200

	xx = np.linspace(0, 20, n_points)

	yy = intercept + alpha*xx 
	yy += np.random.normal(size=n_points)


	pr = Fit(xx, yy, n_breakpoints=0, n_boot=25)
	pr.summary()


def model_comparision_straight_line():

	np.random.seed(0)

	alpha = 10
	intercept = 100

	n_points = 200

	xx = np.linspace(0, 20, n_points)

	yy = intercept + alpha*xx 
	yy += np.random.normal(size=n_points)


	ms = ModelSelection(xx, yy, max_breakpoints=6)


if __name__=="__main__":

	on_data_1()



import numpy as np
import scipy
import math

def get_r_squared(yy, ff, n_params):

	n_data = len(yy)
	yy_mean = np.mean(yy)

	# Calculate residual and total sum of squares
	residual_sum_squares = 0
	total_sum_squares = 0
	for i in range(n_data):
		residual_sum_squares += (yy[i] - ff[i])**2
		total_sum_squares += (yy[i] - yy_mean)**2

	# R Squares
	r_squared = 1 - residual_sum_squares/total_sum_squares

	# Adjusted R squared
	adjusted_r_squared = 1 - (1-r_squared) * (n_data-1) /(n_data - n_params - 1)
	
	return residual_sum_squares, total_sum_squares, r_squared, adjusted_r_squared



import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
import math

import os, sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))

from piecewise_regression.main import Muggeo, Fit


def check_p_values_fit():

    p_counts ={
        "alpha1":0,
        "beta1":0,
        "alpha2":0,
        "const":0
    }

 
    sample_size = 1000

    actuals = {
        "alpha1":1,
        "beta1":3,
        "alpha2":4,
        "const":5
    }

    for seed in range(sample_size):
        print("Working on {} of {} . . . ".format(seed, sample_size))
        np.random.seed(seed)

        xx_bp, yy_bp = generate_data(intercept=actuals["const"], alpha=actuals["alpha1"], beta_1=actuals["beta1"])

        bp_fit = Fit(xx_bp, yy_bp, [5], verbose=False)

        for estimate in ["alpha1", "beta1", "alpha2", "const"]:
            est_ci = bp_fit.best_muggeo.best_fit.estimates[estimate]["confidence_interval"]
            if est_ci[0] <= actuals[estimate] and est_ci[1]>= actuals[estimate]:
                p_counts[estimate] += 1

    print("Percentages in confidence interval:")
    for estimate in ["alpha1", "beta1", "alpha2", "const"]:        
        print("{} : {}".format(estimate, p_counts[estimate]/sample_size))
    print("This should be approximately 95%")    



def generate_data(intercept, alpha, beta_1):

    intercept = 5
    alpha = 1
    beta_1 = 3
    n_points = 50
    noise = 1
    breakpoint_1 = 2

    xx_bp = np.linspace(-9.5, 9.5, n_points)
    
    yy_bp = intercept + alpha*xx_bp + beta_1 * np.maximum(xx_bp - breakpoint_1, 0) + noise*np.random.normal(size=len(xx_bp))


    return xx_bp, yy_bp





if __name__=="__main__":
    check_p_values_fit()



import math
import warnings

import numpy as np
import statsmodels.api as sm
import scipy.stats
import matplotlib.pyplot as plt

try:
	import piecewise_regression.davies as davies 
	import piecewise_regression.r_squared_calc as r_squared_calc
	from piecewise_regression.data_validation import validate_positive_number, validate_boolean, validate_positive_integer, validate_list_of_numbers, validate_non_negative_integer
except:
	import davies
	import r_squared_calc
	from data_validation import validate_positive_number, validate_boolean, validate_positive_integer, validate_list_of_numbers, validate_non_negative_integer


class NextBreakpoints:
	"""
	One iteration of Muggeo's segmented regression algorithm. Gets the next breakpoints. 
	Also calculates interesting statistics. This expects data validation and error 
	handling are done at a higher level.

	:param xx: Data series in x-axis for fitting (same axis as the breakpoints)
	:type xx: list

	:param yy: Data series in y-axis for fitting
	:type yy: list

	:param current_breakpoints: The starting breakpoints for this iteration
	:type current_breakpoints: list

	"""
	def __init__(self,
		xx, # list(float) or numpy(float). REQUIRED. Data series in x-axis
		yy, # list(float) or numpy(float). REQUIRED. Data series in y-axis
		current_breakpoints # list(float) or numpy(float). REQUIRED. Current values of breakpoint positions 
		):

		# Data validation done at a higher level
		self.xx = xx
		self.yy = yy
		self.current_breakpoints = current_breakpoints
		self.n_breakpoints = len(current_breakpoints)

		self.next_breakpoints = None
		self.raw_params = None
		self.covariance_matrix = None

		self.breakpoint_fit()

		# All estimate data saved in dictionary
		self.estimates = {}

		# Don't really need to do this at this point. But it is very quick and nice to have the record of these as we go 
		self.calculate_all_estimates()
		self.calculate_all_standard_errors()
		self.calculate_all_confidence_intervals()
		self.calculate_all_t_stats()

		# R squared etc
		self.residual_sum_squares = None 
		self.total_sum_squares = None
		self.r_squared = None
		self.adjusted_r_squared = None
		self.bic = None

		self.calculate_r_squared()
		self.calculate_bayesian_information_criterion()


	def breakpoint_fit(self):
		"""
		Fit the linear approximation given the current breakpoint guesses. Sets the next breakpoints 
		and the params from the fit. The params are of the form [c, a, beta_hats, gamma_hats]
		"""
		Z = np.array([self.xx])
		# Convert data based on breakpoints

		UU = [(self.xx - bp) * np.heaviside(self.xx - bp, 1) for bp in self.current_breakpoints]
		VV = [np.heaviside(self.xx- bp, 1) for bp in self.current_breakpoints]

		
		Z = np.concatenate((Z, UU, VV))
		Z = Z.T	
		Z = sm.add_constant(Z, has_constant='add')

		results = sm.OLS(endog=self.yy, exog=Z).fit()

		self.raw_params = results.params
		self.covariance_matrix = results.cov_params()

		# First two params are a and c in the line equation
		# Beta hats are the next group of params, same length as the number of breakpoints
		beta_hats = results.params[2:2+len(self.current_breakpoints)]
		# Gamma hats are the last group of params, same length as the number of breakpoints
		gamma_hats = results.params[2+len(self.current_breakpoints):]
		# The next breakpoints are calculated iteratively
		self.next_breakpoints = self.current_breakpoints - gamma_hats/beta_hats


	def calculate_all_estimates(self):
		"""
		Extract estiamtes from the params and saves in the self.estimates dictionary
		"""
		params = self.raw_params
	
		# Extract the exstimates form the correct locations in the params returned from statsmodels
		const_estimate = params[0]
		beta_estimates = params[2:self.n_breakpoints+2]
		breakpoint_estimates = self.next_breakpoints
		
		self.estimates["const"] = {"estimate":const_estimate}
		for bp_i in range(self.n_breakpoints):
			self.estimates["beta{}".format(bp_i+1)] = {"estimate":beta_estimates[bp_i]}
			self.estimates["breakpoint{}".format(bp_i+1)] = {"estimate":breakpoint_estimates[bp_i]}
		# Also calculate alphas
		for alpha_i in range(self.n_breakpoints + 1):
			alpha_estimate = np.sum(params[1:alpha_i+2])
			self.estimates["alpha{}".format(alpha_i+1)] = {"estimate":alpha_estimate}


	def get_alpha_standard_errors(self):
		"""
		Get the standard errors for the alphas (gradients of segments)
		"""
		cov_matrix = self.covariance_matrix

		# Alphas are calculated as the sum of alpha_0 and betas up that part of the regression line 
		# The var of each alpha is the sum of the covariance matrix up to it
		# Removing the intercept column and row. 
		# var(alpha_k) = var(alpha_1) + sum var(beta_j) + 2*sum_{i=1,j=2}^k *cov(alpha, betas))
		# var(alpha_k) = sum_{i, j} cov(alpha and betas)
		alpha_vars = []
		for alpha_n in range(self.n_breakpoints + 1):
			alpha_cov_matrix = cov_matrix[1:alpha_n+2, 1:alpha_n+2]	
			alpha_vars.append(np.sum(alpha_cov_matrix))

		alpha_ses = np.sqrt(alpha_vars)	
		return alpha_ses	


	def get_bp_standard_errors(self):
		"""
		Get the standard errors of the breakpoints. Considering bp = gamma/beta + bp_0, the 
		standard error of the breakpoint estaimtes can be found using the ratio/delta method. 
		See e.g. Muggeo (2003) for clarification
		"""
		cov_matrix = self.covariance_matrix
		params = self.raw_params

		bp_vars = []

		# For each breakpoint, calcaulte the variance of the estimator
		for bp_n in range(self.n_breakpoints):
			beta_index = 2 + bp_n
			gamma_index = 2 + self.n_breakpoints + bp_n

			beta = params[beta_index]
			gamma = params[gamma_index]
			gamma_var = cov_matrix[gamma_index, gamma_index]
			beta_var = cov_matrix[beta_index, beta_index]
			gamma_beta_covar = cov_matrix[beta_index, gamma_index]

			# From Muggeo (2003). The sign before the covariance term is opposite to Muggeo, this is because the gamma is defined with the opposite sign
			# The calculation is equivalent to Muggeos.
			bp_var = (gamma_var + beta_var * (gamma/beta)**2 - 2*(gamma/beta)*gamma_beta_covar)/(beta**2)
			bp_vars.append(bp_var)

		bp_ses = np.sqrt(bp_vars)
		return bp_ses


	def get_const_standard_error(self):
		"""
		Get the constant standard error from the covariance matrix
		""" 
		# Covariance matrix is [c, alpha, betas, gammas]
		# Constant variance is just the top left cell in covariance matrix
		cov_matrix = self.covariance_matrix
		c_var = cov_matrix[0][0]
		return np.sqrt(c_var)

	def get_beta_standard_errors(self):
		"""
		Get the beta estimates standard errors from the covariance matrix
		"""
		# Covariance matrix is [c, alpha, betas, gammas]
		# Beta variances are along the diagonal of the covariance matrix
		cov_matrix = self.covariance_matrix
		cov_diagonal = np.diagonal(cov_matrix)
		beta_vars = cov_diagonal[2:self.n_breakpoints+2]
		return np.sqrt(beta_vars)


	def calculate_all_standard_errors(self):
		"""
		Calculate standard errors for all the variables of interest. 
		Save to the self.estimates dictionary
		"""
		const_ses = self.get_const_standard_error()
		self.estimates["const"]["se"] = const_ses

		beta_ses = self.get_beta_standard_errors()
		bp_ses = self.get_bp_standard_errors()
		for bp_i in range(self.n_breakpoints):
			self.estimates["beta{}".format(bp_i+1)]["se"] = beta_ses[bp_i]
			self.estimates["breakpoint{}".format(bp_i+1)]["se"] = bp_ses[bp_i]
		alpha_ses = self.get_alpha_standard_errors()
		for alpha_i in range(self.n_breakpoints + 1):
			self.estimates["alpha{}".format(alpha_i+1)]["se"] = alpha_ses[alpha_i]


	def calculate_all_confidence_intervals(self):
		"""
		Calculate all confidence intervals, based on t-distribution and standard errors.
		"""	
		# Estimates
		dof = len(self.xx) - 2 - 2*self.n_breakpoints
		t_const = scipy.stats.t.ppf(0.975, dof)

		# Iterate over the estimate dictionary, add confidence intervals to all estaimtors 
		for estimator_name, details in self.estimates.items():
			confidence_interval = (details["estimate"] - t_const*details["se"], details["estimate"] + t_const*details["se"])
			details["confidence_interval"] = confidence_interval

	def calculate_all_t_stats(self):
		"""
		Get t stats for all the estimators
		"""
		dof = len(self.xx) - 2 - 2*self.n_breakpoints
		for estimator_name, details in self.estimates.items():
			# Breakpoint t stats don't make sense
			# Don't exist in the null model - nuisance parameter
			# H_0 isn't bp=0, it's that bp doesn't exist
			if "breakpoint" in estimator_name:
				details["t_stat"] = "-"
				details["p_t"] = "-"
			else:
				t_stat = details["estimate"]/details["se"]
				p_t = scipy.stats.t.sf(np.abs(t_stat), dof)*2
				details["t_stat"] = t_stat
				if "beta" in estimator_name:
					details["p_t"] = "-"
				else:	
					details["p_t"] = p_t		


	def get_predicted_yy(self):
		"""
		Get the model predictions for each of the xx data points
		"""
		params = self.raw_params
		breakpoints = self.next_breakpoints
		# Extract what we need from params etc
		intercept_hat = params[0]
		alpha_hat = params[1]
		beta_hats = params[2:2+len(breakpoints)]

		yy_predicted = []

		yy_predicted = intercept_hat + alpha_hat*self.xx
		for bp_count in range(len(breakpoints)):
			yy_predicted += beta_hats[bp_count] * np.maximum(self.xx - breakpoints[bp_count], 0)
		
		return yy_predicted

	def calculate_r_squared(self):
		"""
		Calculate R squared from the fitted model.
		"""
		yy_predicted = self.get_predicted_yy()
		n_params = 2 * self.n_breakpoints + 2

		self.residual_sum_squares, self.total_sum_squares, self.r_squared, self.adjusted_r_squared = r_squared_calc.get_r_squared(self.yy, yy_predicted, n_params)

	def calculate_bayesian_information_criterion(self):
		"""
		Calculates the Bayesian Information Criterion of the fitted model. 
		Assuming normal noise, uses the standard version for OLS models. 
		This should hold for breakpoint regression models, because the BIC is based on the likelihood of the data
		given the model. That likelihood function doesn't involve the breakpoint values - it just depends on distances 
		of the data to the fitted model predictions. Also depends on the error in the noise term being constant.  
		"""
		n = len(self.xx) # No. data points
		k =  2 + 2*self.n_breakpoints # No. model parameters
		rss = self.residual_sum_squares
		self.bic = n * np.log(rss/n) + k * np.log(n)


class Muggeo:
	"""
	Muggeo's iterative segmented regression method. This is a simple version. 
	Errors are handled at a higher level in the Fit object. See Muggeo (2003) for more information.
	If the breakpoints get too close together, or get outside (or near the edge) of the data range, 
	the algorithm is stopped because this is likely to be a local minimum that is difficult to escape, as well
	as possibly generating errors in the iterative procedue. 

	:param xx: Data series in x-axis for fitting (same axis as the breakpoints)
	:type xx: list of floats

	:param yy: Data series in y-axis for fitting
	:type yy: list of floats

	:param n_breakpoints: The number of breakpoints to fit
	:type n_breakpoints: positive int

	:param start_values: A list of initial guesses for the breakpoints
	:type start_values: list floats

	:param verbose: If True, prints out updates to the terminal
	:type verbose: bool	

	:param max_iterations: How many iterations before stopping if not converged
	:type max_iterations: positive int		

	:param tolerance: How close breakpoints from pervious iterations must be to consider converged. 
	:type tolerance: positive float		

	:param min_distance_between_breakpoints: The minimum allowed distance between breakpoints, as a proportion of the data range.
	:type min_distance_between_breakpoints: positive float

	:param min_distance_between_breakpoints: The minimum allowed distance from the edge of data to a breakpoint, as a proportion of the data range.
	:type min_distance_between_breakpoints: positive float	

	"""

	def __init__(self, 
		xx, # list(float) or numpy(float). REQUIRED. Data series in x-axis
		yy, # list(float) or numpy(float). REQUIRED. Data series in y-axis
		n_breakpoints, # int. REQUIRED. Number of breakpoints
		start_values=None, # list(float) or numpy(float). REQUIRED. Initial guesses for breakpoint positions 
		verbose=False, # Boolean. whether to print progress to terminal. 
		max_iterations=30, # Positive int. Maximum iterations of Muggeo algorithm if not converged
		tolerance=10**-5, # Positive float. If breakpoints change less than the tolerance then the algorithm has converged
		min_distance_between_breakpoints=0.01, # Positive float. The minimum required distance between breakpoints, as a proportion of the data range. 
		min_distance_to_edge=0.02, # Positive float. Minimum distance from edge of data to a breakpoint, as a proportion of the data range. 
		):

		self.verbose = verbose
		if self.verbose:
			print("\nInstantiating Muggeo . . . with start_values = {}".format(start_values))

		# validation is done at a higher level
		self.xx = xx
		self.yy = yy

		self.n_breakpoints = n_breakpoints
		self.min_distance_between_breakpoints = min_distance_between_breakpoints
		self.min_distance_to_edge = min_distance_to_edge

		if start_values is None:
			start_values = self._generate_breakpoints()

		self.start_values = self._validate_start_values(start_values)
		
		self.max_iterations = max_iterations	
		self.tolerance = tolerance
		self.verbose = verbose

		self.stop = False

		if self.n_breakpoints != len(self.start_values):
			raise ValueError("n_breakpoints is not the same as the length of start_values")

		self.fit_history = []
		self.best_fit = None
		self.converged = False
		
		# Records the reason why the algorithm stopped
		self.stop_reason = None

		self.fit()


	def fit(self):
		"""
		Runs the breakpoint iterative procedure
		"""
		if self.verbose:
			print("Running Muggeo's iterative algorithm . . . ")
		while not self.stop:
			# Do the fit
			# Get the current breakpoints. If first fit then use start_vakues, otherwise use fit history
			if len(self.fit_history) == 0:
				current_breakpoints = self.start_values
			else:
				current_breakpoints = self.fit_history[-1].next_breakpoints
			
			try:
				with warnings.catch_warnings():
					warnings.simplefilter("ignore")
					IteratedFit = NextBreakpoints(self.xx, self.yy, current_breakpoints)
					self.fit_history.append(IteratedFit)
					self.stop_or_not()
					if self.verbose:
						print("Next breakpoints are ", IteratedFit.next_breakpoints)
			except Exception as e:
				self.stop=True
				self.stop_reason = "Error encountered: " + str(e)

		# Select the best fit from the fit history by finding the smallest rss
		if len(self.fit_history) > 0:
			self.best_fit = min(self.fit_history, key=lambda x:x.residual_sum_squares)

	def stop_or_not(self):
		"""
		Test to see if the iterative procedure should stop. 
		Stop if it's converged, if max_iterations reached, of the breakpoints fall into values that
		are too close together or outside of the allowed rance.
		"""
		# Stop if maximum iterations reached
		if len(self.fit_history)>self.max_iterations:
			self.stop_reason = "Algorithm was stopped as maximum iterations reached"
			self.stop = True

		if not self._are_breakpoint_values_far_apart(self.fit_history[-1].next_breakpoints):
			self.stop_reason = "Breakpoint values too close together"
			self.stop = True

		if not self._are_breakpoint_values_within_range(self.fit_history[-1].next_breakpoints):
			self.stop_reason = "Breakpoint values outside range"
			self.stop=True

		# Stop if tolerance reached - small change between this and last breakpoints
		if len(self.fit_history) > 1:
			breakpoint_differences = self.fit_history[-2].next_breakpoints - self.fit_history[-1].next_breakpoints
			if np.max(np.abs(breakpoint_differences)) <= self.tolerance:
				self.stop_reason = "Algorithm converged on breakpoint values"
				self.stop = True
				self.converged = True

		# Stop if the algorithm is iterating back to previous values, within tolerance
		if len(self.fit_history) > 2:
			breakpoint_two_step_differences = self.fit_history[-3].next_breakpoints - self.fit_history[-1].next_breakpoints
			if np.max(np.abs(breakpoint_two_step_differences)) <= self.tolerance:
				self.stop_reason = "Algorithm converged on breakpoint values"
				self.stop = True
				self.converged = True

	def _validate_start_values(self, start_values):
		"""
		Validate the breakpoint start_values.  
		Should be a list of numbers, or a numpy list of numbers.
		They should also be not too close to the edge of the data, and not too close together - to avoid the Muggeo algorithm diverging
	
		:param start_values: A list of initial guesses for the breakpoints
		:type start_values: list floats

		"""

		start_values = validate_list_of_numbers(start_values, "start_values", min_length=1)

		if not self._are_breakpoint_values_within_range(start_values):
			self.stop = True
			self.stop_reason = "start_values are not within allowed range. Try changing min_distance_to_edge"
			
		if not self._are_breakpoint_values_far_apart(start_values):
			self.stop = True
			self.stop_reason = "start_values are too close together. Try changing min_distance_between_breakpoints"
		return start_values


	def _are_breakpoint_values_within_range(self, breakpoints):
		"""
		Check that breakpoints are self.min_distance_to_edge * range away for the edge of the data in x

		:param breakpoints: A list of breakpoints
		:type breakpoints: list floats

		"""
		min_allowed_bp = np.quantile(self.xx, self.min_distance_to_edge)
		max_allowed_bp = np.quantile(self.xx, 1-self.min_distance_to_edge)

		for bp in breakpoints:
			if bp <= min_allowed_bp or bp >= max_allowed_bp:
				return False				
		return True

	def _are_breakpoint_values_far_apart(self, breakpoints):
		"""
		Check if breakpoint values are self.min_distance_between_breakpoints*range away from each other
		
		:param breakpoints: A list of breakpoints
		:type breakpoints: list floats

		"""
		min_distance = np.diff(np.sort(breakpoints))

		# numpy ptp gives the range of the data, closeness realtive to the range
		min_distance_allowed = self.min_distance_between_breakpoints * np.ptp(self.xx)

		if (min_distance <= min_distance_allowed).any():
			return False
		return True

	def _generate_breakpoints(self):
		"""
		Randomly generate some breakpoint values
		"""
		# Get breakpoints within allowed range
		min_allowed_bp = np.quantile(self.xx, self.min_distance_to_edge)
		max_allowed_bp = np.quantile(self.xx, 1-self.min_distance_to_edge)
		start_values = np.random.uniform(low=min_allowed_bp, high=max_allowed_bp, size=self.n_breakpoints)
		if self.verbose:
			print("Generating some random breakpoints: ", start_values)
		return start_values


class Fit:
	"""
	Fit a peicewise (segmented) regression model to data. 
	Uses bootstrap restarting to avoid local minima. 
	Requires either n_breakpoints of start_values. 
	if no start_vaues are given, they are instead uniformly randomly generated across range of data.
	This is the main user facing object and input data is validated mainly at this level. 

	:param xx: Data series in x-axis for fitting (same axis as the breakpoints).
	:type xx: list of floats

	:param yy: Data series in y-axis for fitting.
	:type yy: list of floats

	:param n_breakpoints: The number of breakpoints to fit.
	:type n_breakpoints: positive int

	:param start_values: A list of initial guesses for the breakpoints.
	:type start_values: list floats

	:param n_boot: How many times to run the bootstrap restarting procedure. Set to zero for no bootstrap restarting. 
	:type n_boot: non-negative int

	:param verbose: If True, prints out updates to the terminal.
	:type verbose: bool	

	:param max_iterations: How many iterations before stopping if not converged, in the Muggeo iterative procedure.
	:type max_iterations: positive int		

	:param tolerance: How close breakpoints from pervious iterations must be to consider converged. 
	:type tolerance: positive float		

	:param min_distance_between_breakpoints: The minimum allowed distance between breakpoints, as a proportion of the data range.
	:type min_distance_between_breakpoints: positive float

	:param min_distance_between_breakpoints: The minimum allowed distance from the edge of data to a breakpoint, as a proportion of the data range.
	:type min_distance_between_breakpoints: positive float	

	"""
	def __init__(self, 
		xx, # list(float) or numpy(float). REQUIRED Data series in x-axis
		yy, # list(float) or numpy(float). REQUIRED Data series in y-axis
		start_values=None, # list(float) or numpy(float). Initial guesses for breakpoint positions 
		n_breakpoints=None, # int. If not start_values, the number of breakpoints to fit. REQUIRED if no start_values 
		n_boot=20, # Positive int. The number of times to run the bootstrap restarting. n_boot=0 runs the Muggeo algorithm with no bootstrap
		verbose=False, # Boolean. whether to print progress to terminal. 
		max_iterations=30, # Positive int. Maximum iterations of Muggeo algorithm if not converged
		tolerance=10**-5, # Positive float. If breakpoints change less than the tolerance then the algorithm has converged
		min_distance_between_breakpoints=0.01, # Positive float. The minimum required distance between breakpoints, as a proportion of the data range. 
		min_distance_to_edge=0.02, # Positive float. Minimum distance from edge of data to a breakpoint, as a proportion of the data range. 
		):


		self.verbose = validate_boolean(verbose, "verbose")
		
		if self.verbose:
			print("\nInstantiating Fit . . . ")

		# Validate all input data
		self.xx = validate_list_of_numbers(xx, "xx", min_length=3)
		self.yy = validate_list_of_numbers(yy, "yy", min_length=3)

		if len(self.yy) != len(self.xx):
			raise ValueError("x and y data series must be the same size len(xx)={}, len(yy)={}".format(len(self.xx), len(self.yy)))

		self.n_boot = validate_non_negative_integer(n_boot, "n_boot")
		self.max_iterations = validate_positive_integer(max_iterations, "max_iterations")
		self.tolerance = validate_positive_number(tolerance, "tolerance")
		self.min_distance_between_breakpoints = validate_positive_number(min_distance_between_breakpoints, "min_distance_between_breakpoints")
		self.min_distance_to_edge = validate_positive_number(min_distance_to_edge, "min_distance_to_edge")

		# We need either start_values or n_breakpoints
		if n_breakpoints is None:
			self.n_breakpoints = None
		else:
			self.n_breakpoints = validate_positive_integer(n_breakpoints, "n_breakpoints")

		if start_values is None:
			self.start_values = None
		else:
			self.start_values = validate_list_of_numbers(start_values, "start_values", min_length=1)
			self.n_breakpoints = len(self.start_values)

		if start_values is None and n_breakpoints is None:
			raise ValueError("The Fit algorithm requires either the start_values or n_breakpoints")								

		self.bootstrap_history = []
		self.best_muggeo = None
		self.stop = False

		self.bootstrap_restarting()

		self.davies = davies.davies_test(self.xx, self.yy)

	def get_results(self):
		"""
		Return a small dictionary with key results form the fit.
		Useful for using this code in a larger analysis. E.g. ModelSelection
		"""
		results = {
			"davies":self.davies,
		}

		if self.best_muggeo:
			results["estimates"] = self.best_muggeo.best_fit.estimates
			results["bic"] = self.best_muggeo.best_fit.bic
			results["rss"] = self.best_muggeo.best_fit.residual_sum_squares
			results["converged"] = True
		else:
			results["converged"] = False
			results["estimates"] = None
			results["bic"] = None
			results["rss"] = None
		return results

	def bootstrap_restarting(self):
		"""
		The main fitting algorithm. Begins by doing a fit based on Muggeo's algorithm. 
		if n_boot = 0 we stop there. Otherwise we do some bootstrap restarting.  
		Bootstrap Restarting escapes local minima.
		Each bootstrap restart: 
		- We take the best current breakpoints, and get new data by running a non-parametric bootstrap by resampling data. 
		- Then run a Muggeo fit on the new data and best current breakpoints. This gives new breakpoint values. 
		- Then run a Muggeo fit again with the original data and these new breakpoint values. 
		- Throughout, keep track of the history of fits and the best_muggeo fit that converged - defined as the lowest residual sum of squares.
		"""			
		muggeo_fit = Muggeo(xx=self.xx, yy=self.yy, start_values=self.start_values, n_breakpoints = self.n_breakpoints,
			max_iterations=self.max_iterations, tolerance = self.tolerance, verbose=self.verbose, 
			min_distance_between_breakpoints=self.min_distance_between_breakpoints, 
			min_distance_to_edge=self.min_distance_to_edge)

		self.bootstrap_history.append(muggeo_fit)

		# best_muggeo is the best converged muggeo
		if muggeo_fit.converged:
			self.best_muggeo = muggeo_fit
		
		# Iterate bootstraps
		for i in range(self.n_boot):

			# Best breakpoints are either from best converged muggeo so far, or start values, or randomly generated
			if self.best_muggeo:
				best_bps = self.best_muggeo.best_fit.next_breakpoints
			else:
				best_bps = self.start_values
				
			# Get some new breakpoint values from a bootstrapped fit
			# Non parametric bootstrap by resampling from data
			xx_boot, yy_boot = self.bootstrap_data(self.xx, self.yy)
			bootstrap_fit = Muggeo(xx_boot, yy_boot, start_values=best_bps, n_breakpoints = self.n_breakpoints,
				max_iterations=self.max_iterations, tolerance = self.tolerance, verbose=self.verbose, 
				min_distance_between_breakpoints=self.min_distance_between_breakpoints, 
				min_distance_to_edge=self.min_distance_to_edge)
			if bootstrap_fit.converged:
				bootstrap_bps = bootstrap_fit.best_fit.next_breakpoints
			else:
				# Give it something - even though these breakpoints are already run
				# prefer this to using breaks in the for loop
				bootstrap_bps = best_bps

			# Do a new fit with the new breakpoint values 
			next_muggeo = Muggeo(self.xx, self.yy, start_values=bootstrap_bps, n_breakpoints = self.n_breakpoints,
				max_iterations=self.max_iterations, tolerance = self.tolerance, verbose=self.verbose, 
				min_distance_between_breakpoints=self.min_distance_between_breakpoints, 
				min_distance_to_edge=self.min_distance_to_edge)
			self.bootstrap_history.append(next_muggeo)

			# If we get a converged answer, see if this new fit is the best or not
			if next_muggeo.converged:
				# If there is already a converged best_muggeo, see if this one is better
				if self.best_muggeo:
					if next_muggeo.best_fit.residual_sum_squares < self.best_muggeo.best_fit.residual_sum_squares:
						self.best_muggeo = next_muggeo
				# If there is not already a converged best_muggeo, use this fit instead
				else:
					self.best_muggeo = next_muggeo

	def bootstrap_data(self, xx, yy):
		"""
		Non parametric bootstrap, randomly sample data points with replacement. 
		Return bootstrapped data of same length as oriignal data.

		:param xx: Data series in x-axis.
		:type xx: list of floats

		:param yy: Data series in y-axis.
		:type yy: list of floats

		"""
		n = len(xx)
		# Get bootstrap samples as array index locations
		boot_indices = np.random.choice(n, size=n, replace=True)

		xx_boot = xx[boot_indices]
		yy_boot = yy[boot_indices]
		return xx_boot, yy_boot

	def plot_data(self, **kwargs):
		"""
		Plot the data as a scatter plot. 
		Passes any kwargs to the matplotlib scatter function, e.g. color="red".

		"""
		plt.scatter(self.xx, self.yy, **kwargs)

	def plot_fit(self, **kwargs):
		"""
		Plot the fitted model as a series of straight lines. 
		Passes any kwargs to the matplotlib plot function, e.g. color="red".

		"""
		if not self.best_muggeo:
			print("Algorithm didn't converge. No fit to plot.")
		else:
			# Get the final results from the fitted model variables
			# Params are in terms of [intercept, alpha, betas, gammas]
			final_params = self.best_muggeo.best_fit.raw_params
			breakpoints = self.best_muggeo.best_fit.next_breakpoints
			
			# Extract what we need from params etc
			intercept_hat = final_params[0]
			alpha_hat = final_params[1]
			beta_hats = final_params[2:2+len(breakpoints)]
			
			xx_plot = np.linspace(min(self.xx), max(self.xx), 100)

			# Build the fit plot segment by segment. Betas are defined as difference in gradient from previous section
			yy_plot = intercept_hat + alpha_hat*xx_plot
			for bp_count in range(len(breakpoints)):
				yy_plot += beta_hats[bp_count] * np.maximum(xx_plot - breakpoints[bp_count], 0)

			plt.plot(xx_plot, yy_plot, **kwargs)

	def plot_breakpoints(self, **kwargs):
		"""
		Plot the breakpoint locations as vertical lines. 
		Passes kwargs to the matplotlib function, e.g. color="red".

		"""
		if not self.best_muggeo:
			print("Algorithm didn't converge. No breakpoints to plot")
		else:
			breakpoints = self.best_muggeo.best_fit.next_breakpoints			
			
			for bp in breakpoints:
				plt.axvline(bp, **kwargs)

	def plot_breakpoint_confidence_intervals(self, **kwargs):
		"""
		Plot the breakpoint confidence intervals as vertical shaded regions.
		Passes kwargs to the matplotlib function, e.g. color="red".
		
		"""
	
		if not self.best_muggeo:
			print("Algorithm didn't converge. No breakpoint estimates to plot")
		else:			
			estimates = self.best_muggeo.best_fit.estimates			
			
			for bp_i in range(self.best_muggeo.n_breakpoints):
				bp_ci = estimates["breakpoint{}".format(bp_i+1)]["confidence_interval"]
				plt.axvspan(bp_ci[0], bp_ci[1], alpha=0.1)

	def plot_best_muggeo_breakpoint_history(self, **kwargs):
		"""
		Plot the history of the breakpoints as they iterate. 
		History of the best_muggeo fit.

		"""
		if not self.best_muggeo:
			print("Algorithm didn't converge. No meaningful history to plot")
		else:
			# Get the data from the best_muggeo in a form for plotting
			breakpoint_history = [self.best_muggeo.start_values]
			for fit_details in self.best_muggeo.fit_history:
				breakpoint_history.append(fit_details.next_breakpoints)
			breakpoint_history_series = zip(*breakpoint_history)
			
			# Plot a history for each breakpoint 
			count = 0
			for bh in breakpoint_history_series:
				count += 1
				plt.plot(range(0, len(bh)), bh, label="Breakpoint {}".format(count), **kwargs)
				plt.xlabel("Muggeo Iteration")
				plt.ylabel("Breakpoint")

	def plot_bootstrap_restarting_history(self, **kwargs):
		"""
		Plot the history of the breakpoint values as they iterate. 
		History of the bootstrap restarting procedure.

		"""
		if not self.best_muggeo.converged:
			print("Algorithm didn't converge. Plotting breakpoint history anyway")

		# Get the data from the best_muggeo in a form for plotting
		
		breakpoint_history = [self.start_values]

		for muggeo_fit in self.bootstrap_history:
			if muggeo_fit.best_fit:
				breakpoint_history.append(muggeo_fit.best_fit.next_breakpoints)
			else:
				breakpoint_history.append(None)

		breakpoint_history_series = zip(*breakpoint_history)
		
		# Plot a history for each breakpoint 
		count = 0
		for bh in breakpoint_history_series:
			count += 1
			plt.plot(range(0, len(bh)), bh, label="Breakpoint {}".format(count), **kwargs)
			plt.xlabel("Bootstrap Iteration")
			plt.ylabel("Breakpoint")

	def plot_bootstrap_restarting_rss_history(self, **kwargs):
		"""
		Plot the history of the residual sum of squares. 
		History of the bootstrap restarting algorithm.

		"""
		if not self.best_muggeo:
			print("Algorithm didn't converge. Plotting rss history anyway")

		# Get the data from the best_muggeo in a form for plotting
		
		rss_history = []

		for muggeo_fit in self.bootstrap_history:
			if muggeo_fit.best_fit:
				rss_history.append(muggeo_fit.best_fit.residual_sum_squares)
			else:
				rss_history.append(None)

		plt.plot(range(1, len(rss_history) + 1), rss_history, **kwargs)
		plt.xlabel("Bootstrap Iteration")
		plt.ylabel("Residual Sum of Squares")

	def plot(self):
		"""
		Plot the full fit including the data, fitted model, breakpoint positions and breakpoint confidence intervals.
		Doesn't allow control of matplotlib kwargs for style changes.
		"""
		self.plot_data()
		self.plot_fit()
		self.plot_breakpoints()
		self.plot_breakpoint_confidence_intervals()

	def summary(self):
		"""
		Print a summary of the best fit, along the lines of the summary given by python's statsmodels OLS fit.
		
		"""

		if not self.best_muggeo:
			print("Algorithm did not converge. Try different n_breakpoints, different start_values, or start_values=None")

			print("Summary of why the algorithm did not converge:")
			run_count = 1
			for muggeo_fit in self.bootstrap_history:
				print("Run {}: {}".format(run_count, muggeo_fit.stop_reason))
				run_count += 1
		else:
			header = "\n{:^70}\n".format("Breakpoint Regression Results")

			line_length=100
			double_line = "=" * line_length + "\n"
			single_line = "-" * line_length + "\n"

			# Overview
			n_obs = len(self.xx)
			n_model_params = 2 + 2*self.n_breakpoints
			dof = n_obs - n_model_params
			no_obs_text = "{:<20} {:>20}\n".format("No. Observations", n_obs)
			no_model_parameters_text = "{:<20} {:>20}\n".format("No. Model Parameters", n_model_params)
			dof_text = "{:<20} {:>20}\n".format("Degrees of Freedom", dof)
			rss_text = "{:<20} {:>20.6}\n".format("Res. Sum of Squares", self.best_muggeo.best_fit.residual_sum_squares)
			tss_text = "{:<20} {:>20.6}\n".format("Total Sum of Squares", self.best_muggeo.best_fit.total_sum_squares)
			r_2_text = "{:<20} {:>20.6f}\n".format("R Squared", self.best_muggeo.best_fit.r_squared)
			adj_r_2_text = "{:<20} {:>20.6f}\n".format("Adjusted R Squared", self.best_muggeo.best_fit.adjusted_r_squared)
			converged_text = "{:<20} {:>20s}\n".format("Converged: ", str(self.best_muggeo.converged))

			overview = double_line + no_obs_text + no_model_parameters_text + dof_text + rss_text + tss_text + r_2_text + adj_r_2_text + converged_text + double_line

			# Table of results

			table_header_template = "{:<15} {:>12} {:>12} {:>12} {:>12} {:>12} {:>12}\n"

			table_header = table_header_template.format("", "Estimate", "Std Err", "t", "P>|t|", "[0.025", "0.975]")
			#print ("{:<8} {:<15} {:<10}".format( name, age, perc))

			table_row_template = "{:<15} {:>12.6} {:>12.3} {:>12.5} {:>12.3} {:>12.5} {:>12.5}\n"

			table_contents = ""

			beta_names = ["beta{}".format(i+1) for i in range(self.n_breakpoints)]
			bp_names = ["breakpoint{}".format(i+1) for i in range(self.n_breakpoints)]

			model_estimator_names = ["const", "alpha1"] + beta_names + bp_names

			estimates = self.best_muggeo.best_fit.estimates

			for est_name in model_estimator_names:
				estimator_row = table_row_template.format(est_name, estimates[est_name]["estimate"], estimates[est_name]["se"], 
				estimates[est_name]["t_stat"], estimates[est_name]["p_t"], 
				estimates[est_name]["confidence_interval"][0], estimates[est_name]["confidence_interval"][1])
				table_contents += estimator_row

			table_contents += single_line

			table_contents += "These alphas(gradients of segments) are estimated from betas(change in gradient)\n"

			alpha_names = ["alpha{}".format(alpha_i+1) for alpha_i in range(1, self.n_breakpoints+1)]

			table_contents += single_line

			for est_name in alpha_names:
				estimator_row = table_row_template.format(est_name, estimates[est_name]["estimate"], estimates[est_name]["se"], 
				estimates[est_name]["t_stat"], estimates[est_name]["p_t"], 
				estimates[est_name]["confidence_interval"][0], estimates[est_name]["confidence_interval"][1])
				table_contents += estimator_row


			table_contents += double_line

			table = double_line + table_header + single_line + table_contents

			print(header + overview + table)

			print("Davies test for existence of at least 1 breakpoint: p={:.6} (e.g. p<0.05 means reject null hypothesis of no breakpoints at 5% significance)".format(self.davies))
			print("\n\n")


if __name__=="__main__":


	pass



import numpy as np
import scipy
import math

def get_test_statistic(xx_davies, yy_davies, theta):
	"""
	Compute a test statistic for the Davies test for the p-value of existence of a breakpoint.
	Based on Davies(1987) "Hypothesis Testing when a nuisance parameter is present only under the alternative".
	All the variables in this function are as named and described in that paper.

	:param xx_davies: Data series in x-axis (same axis as the breakpoints).
	:type xx_davies: list of floats

	:param yy_davies: Data series in y-axis.
	:type yy_davies: list of floats

	:param theta: A test value from within the range of data in xx
	:type theta: float

	"""
	n = len(xx_davies)
	s_0 = 0
	s_1 = 0
	s_2 = 0
	s_3 = 0
	s_4 = 0

	for x in xx_davies:
		s_0 += x**2

		if x > theta:
			s_1 += x*(x-theta)
			s_3 += x - theta
	
		elif x < theta:
			s_2 += x*(x-theta)
			s_4 += x - theta

	a_hat = np.sum(yy_davies)/n
	b_hat = np.sum(xx_davies*yy_davies)/s_0
	V = s_1*s_2/s_0 + s_3*s_4/n
	S = 0
	for i in range(n):
		if xx_davies[i] > theta:
			S += (yy_davies[i] - a_hat - b_hat*xx_davies[i])*(xx_davies[i] - theta)
	S = S/(np.sqrt(np.abs(V)))
	return S


def davies_test(xx, yy, k=10, alternative="two_sided"):
	"""
	Significance test for the existence of a breakpoint
	Null hypothesis is that there is no breakpoint, or that the change in gradient is zero.
	Alternative hypothesis is that there is a breakpoint, with a non-zero change in gradient.
	The change is gradient is a function of the breakpoint position.
	The breakpoint posiition is a nuisannce parameter that only exists in the alternative hypothesis.
	Based on Davies (1987), "Hypothesis Testing when a nuisance parameter is present only under the alternative".

	:param xx: Data series in x-axis (same axis as the breakpoints).
	:type xx: list of floats

	:param yy: Data series in y-axis.
	:type yy: list of floats

	:param k: A control parameter that determines the number of points to consider within the xx range.
	:type k: int

	:param alternative: Whether to consider a two-sided hypothesis test, or a one sided test with change of gradient greater or less than zero. For existence of a breakpoint, use "two-sided".
	:type alternative: str. One of "two_sided", "less", "greater"	

	"""
	# Centre the x values - makes no difference to existence of a breakpoint
	# The Davies test has this as an assumption
	xx_davies = xx - np.mean(xx)
	yy_davies = yy
	
	# As in Muggeo's R package "segmented", cut from second to second to last data point
	# Need more data in the xx than in [L,U] for the test to work 
	L = xx_davies[1]
	U = xx_davies[-2]

	# More thetas is better
	thetas = np.linspace(L, U, k)
	# For each value of theta, compute a test statistic
	test_stats = []
	for theta in thetas:
		test_stat = get_test_statistic(xx_davies, yy_davies, theta)
		test_stats.append(test_stat)
	if alternative == "two_sided":
		# Two sided test, M as defined by Davies
		M = np.max(np.abs(test_stats))
	elif alternative == "less":
		M = np.abs(np.min(test_stats))
	elif alternative == "greater":
		M = np.max(test_stats)

	# Use formulas from Davies
	V = 0 
	for i in range(len(thetas)-1):
		V += np.abs(test_stats[i+1]-test_stats[i])

	p = scipy.stats.norm.cdf(-M) + V * np.exp(-.5*M**2) * 1/(np.sqrt(8*math.pi))

	if alternative == "two_sided":
		return p*2
	else:
		return p



import numpy as np

def validate_boolean(var, var_name):
	if isinstance(var, bool):
		return var
	else:
		raise ValueError("{} must be a Boolean: True or False".format(var_name))

def validate_positive_integer(var, var_name):
	if isinstance(var, int) and var>0:
		return var
	else:
		raise ValueError("{} must be a positive Integer".format(var_name))

def validate_non_negative_integer(var, var_name):
	if isinstance(var, int) and var>=0:
		return var
	else:
		raise ValueError("{} must be a positive Integer".format(var_name))

def validate_positive_number(var, var_name):
	if (isinstance(var, float) or isinstance(var, int)) and var>0:
		return var
	else:
		raise ValueError("{} must be a Float".format(var_name))

def validate_list_of_numbers(var, var_name, min_length):
	"""
	Allowed types:
		List of integers of floats
		Numpy array of integers or floats
	"""
	value_error_text = "{} must be a list of numbers with minimum length {}".format(var_name, min_length)
	# If its a list, convert it to a numpy array
	if isinstance(var, list):
		var = np.array(var)

	# If its not a numpy array at this point, raise a value error		
	if not isinstance(var, np.ndarray):
		raise ValueError(value_error_text)

	# Check the array has numebrs in it
	if not np.issubdtype(var.dtype, np.number):
		raise ValueError(value_error_text)

	if len(var) < min_length:
		raise ValueError(value_error_text)

	return var





import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
import math

import os, sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))

from piecewise_regression.main import Muggeo, Fit

def check_p_values():

    p_count = 0
    
    sample_size = 1000

    actual_bp = 2


    for seed in range(sample_size):
        print("Working on {} of {} . . . ".format(seed, sample_size))
        np.random.seed(seed)

        xx_bp, yy_bp = generate_data(actual_bp)

        bp_fit = Muggeo(xx_bp, yy_bp, [5], verbose=False)

        bp_ci = bp_fit.best_fit.estimates["breakpoint1"]["confidence_interval"]
        if bp_ci[0] <= actual_bp and bp_ci[1]>= actual_bp:
            p_count += 1  

    print(p_count)

    print("{} of {} estimates were within the confidence interval.".format(p_count, sample_size))
    print("This should be approximately 95%")

def check_p_values_fit():

    p_count = 0
    
    sample_size = 100

    actual_bp = 2


    for seed in range(sample_size):
        print("Working on {} of {} . . . ".format(seed, sample_size))
        np.random.seed(seed)

        xx_bp, yy_bp = generate_data(actual_bp)

        bp_fit = Fit(xx_bp, yy_bp, [5], verbose=False)

        bp_ci = bp_fit.best_muggeo.best_fit.estimates["breakpoint1"]["confidence_interval"]
        if bp_ci[0] <= actual_bp and bp_ci[1]>= actual_bp:
            p_count += 1  

    print(p_count)

    print("{} of {} estimates were within the confidence interval.".format(p_count, sample_size))
    print("This should be approximately 95%")    



def generate_data(breakpoint_1):

    intercept = 5
    alpha = 1
    beta_1 = 3
    n_points = 50
    noise = 1

    xx_bp = np.linspace(-9.5, 9.5, n_points)
    
    yy_bp = intercept + alpha*xx_bp + beta_1 * np.maximum(xx_bp - breakpoint_1, 0) + noise*np.random.normal(size=len(xx_bp))


    return xx_bp, yy_bp





if __name__=="__main__":
    check_p_values_fit()



import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
import math

import os, sys
sys.path.insert(1, os.path.join(sys.path[0], '..'))

from piecewise_regression.davies import davies_test

def check_p_values():

	p_count_20 = 0
	p_count_5 = 0 
	p_count_2 = 0
	p_count_1 = 0

	p_values = []

	sample_size = 1000

	for seed in range(sample_size):
		np.random.seed(seed)

		xx_bp, yy_bp = generate_data()
		theta = 0

		p = davies_test(xx_bp, yy_bp)

		if p<0.2:
			p_count_20 += 1
		if p<0.05:
			p_count_5 +=1
		if p<0.02:
				p_count_2 += 1
		if p<0.01:
			p_count_1 +=1

		p_values.append(p)

	#plt.hist(p_values)
	#plt.show()


	print("P values from empirical testing, along with excpetd p-values")
	print("p-values \t\t\t0.2  \t0.05\t0.02\t0.01")
	print("fraction less than \t{:.3f}\t{:.3f}\t{:.3f}\t{:.3f}".format(p_count_20/sample_size, p_count_5/sample_size,
		p_count_2/sample_size,p_count_1/sample_size))

	print("The empirical fractions should be slghtly less than the expected p values")


def check_p_values_less():

	p_count_20 = 0
	p_count_5 = 0 
	p_count_2 = 0
	p_count_1 = 0

	p_values = []

	sample_size = 1000

	for seed in range(sample_size):
		np.random.seed(seed)

		xx_bp, yy_bp = generate_data()
		theta = 0

		p = davies_test(xx_bp, yy_bp, alternative="less")

		if p<0.2:
			p_count_20 += 1
		if p<0.05:
			p_count_5 +=1
		if p<0.02:
				p_count_2 += 1
		if p<0.01:
			p_count_1 +=1

		p_values.append(p)

	#plt.hist(p_values)
	#plt.show()


	print("P values from empirical testing, along with excpetd p-values")
	print("p-values \t\t\t0.2  \t0.05\t0.02\t0.01")
	print("fraction less than \t{:.3f}\t{:.3f}\t{:.3f}\t{:.3f}".format(p_count_20/sample_size, p_count_5/sample_size,
		p_count_2/sample_size,p_count_1/sample_size))

	print("The empirical fractions should be slightly less than the expected p values")


def generate_data():

	intercept = 5
	alpha = 1
	beta_1 = 0
	breakpoint_1 = 2
	n_points = 20

	xx_bp_1 = list(np.linspace(-9.5, 9.5, n_points))
	xx_bp_2 = list(np.linspace(0,9.5,n_points))
	xx_bp = np.array(xx_bp_1 + xx_bp_2)
	#xx_bp = np.array(xx_bp_1)

	yy_bp = intercept + alpha*xx_bp + beta_1 * np.maximum(xx_bp - breakpoint_1, 0) + np.random.normal(size=len(xx_bp))

	return xx_bp, yy_bp


if __name__=="__main__":
	check_p_values()
	check_p_values_less()


import pathlib
import setuptools

# The directory containing this file
HERE = pathlib.Path(__file__).parent

def readme():
    with open('docs/README.rst') as f:
        return f.read()

# This call to setup() does all the work
setuptools.setup(
    name="piecewise-regression",
    version="1.0.4",
    description="piecewise (segmented) regression in python",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/chasmani/piecewise-regression",
    author="Charlie Pilgrim",
    author_email="pilgrimcharlie2@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
    ],
    packages=setuptools.find_packages(),
    install_requires=["numpy", "matplotlib", "scipy", "statsmodels"],
)